@@ -8,7 +8,6 @@ extern "C" {
 #  error "this header requires Py_BUILD_CORE define"
 #endif
 
-PyAPI_FUNC(PyObject *) _PySys_GetAttr(PyThreadState *, PyObject *); /* unused */
 PyAPI_FUNC(int) _PySys_GetOptionalAttr(PyObject *, PyObject **);
 PyAPI_FUNC(int) _PySys_GetOptionalAttrString(const char *, PyObject **);
 PyAPI_FUNC(PyObject *) _PySys_GetRequiredAttr(PyObject *);
@@ -17,21 +16,16 @@ PyAPI_FUNC(PyObject *) _PySys_GetRequiredAttrString(const char *);
 // Export for '_pickle' shared extension
 PyAPI_FUNC(size_t) _PySys_GetSizeOf(PyObject *);
 
-extern int _PySys_Audit(
-    PyThreadState *tstate,
-    const char *event,
-    const char *argFormat,
-    ...);
-
-// _PySys_ClearAuditHooks() must not be exported: use extern rather than
-// PyAPI_FUNC(). We want minimal exposure of this function.
-extern void _PySys_ClearAuditHooks(PyThreadState *tstate);
-
 extern int _PySys_SetAttr(PyObject *, PyObject *);
 
 extern int _PySys_ClearAttrString(PyInterpreterState *interp,
                                   const char *name, int verbose);
 
+extern int _PySys_SetFlagObj(Py_ssize_t pos, PyObject *new_value);
+extern int _PySys_SetIntMaxStrDigits(int maxdigits);
+
+extern int _PySysRemoteDebug_SendExec(int pid, int tid, const char *debugger_script_path);
+
 #ifdef __cplusplus
 }
 #endif
